defmodule Reactivity.Registry do
  @doc """

  The Registry is responsible for keping track of names of signals. That is,
  derived signals and internal signals.

  """
  use GenServer

  require Logger

  #############
  # GenServer #
  #############

  def start_link(args \\ []) do
    GenServer.start_link(__MODULE__, args, name: __MODULE__)
  end

  def init(_args) do
    table = :ets.new(:sources, [:named_table, :set, :protected])
    {:ok, %{:table => table, :subs => MapSet.new()}}
  end

  #############
  # Callbacks #
  #############

  def handle_cast(m, state) do
    Logger.debug "Cast: #{inspect m}"
    {:noreply, state}
  end


  def handle_call({:insert, source, name}, _from, %{:table => t} = state) do
    :ets.insert(t, {name, source})
    publish_new_signal(source, name, state)
    {:reply, :ok, state}
  end

  def handle_call({:remove, name}, _from, %{:table => t} = state) do
    :ets.delete(t, name)
    {:reply, :ok, state}
  end

  def handle_call({:get, name}, _from, %{:table => t} = state) do
    case :ets.lookup(t, name) do
      [{^name, val}] -> {:reply, {:ok, val}, state}
      []             -> {:reply, {:error, "not found"}, state}
    end
  end

  def handle_call({:subscribe, pid}, _from, %{:subs => ss} = state) do
    Logger.debug "Adding subscription for #{inspect pid}"
    {:reply, :ok, %{state | :subs => MapSet.put(ss, pid)}}
  end

  def handle_call({:unsubscribe, pid}, _from, %{:subs => ss} = state) do
    {:reply, :ok, %{state | :subs => MapSet.delete(ss, pid)}}
  end

  def handle_call(m, from, state) do
    Logger.debug "Call: #{inspect m} from #{inspect from}"
    {:reply, :ok, state}
  end

  def handle_info(m, state) do
    Logger.debug "Info: #{inspect m}"
    {:noreply, state}
  end

  ###########
  # Private #
  ###########

  defp publish_new_signal(source, name, %{:subs => ss} = state) do
    ss
    |> Enum.map(fn(sub) -> send(sub, {:new_signal, source, name}) end)
  end

  #############
  # Interface #
  #############

  @doc """
  Adds a source signal to the registry.
  """
  def add_source(source, name) do
    GenServer.call(__MODULE__, {:insert, source, name})
  end

  @doc """
  Removes a source from the registry by its name.
  """
  def remove_source(name) do
    GenServer.call(__MODULE__, {:remove, name})
  end

  @doc """
  Gets a source signal by means of its name.
  """
  def get_source(name) do
    GenServer.call(__MODULE__, {:get, name})
  end

  @doc """
  Subscribe to events from the signal registry. When a new signal is added, this is sent to the subscribers.
  """
  def subscribe(pid) do
    GenServer.call(__MODULE__, {:subscribe, pid})
  end

  @doc """
  Unsubscribe a given pid from the registry events.
  """
  def unsubscribe(pid) do
    GenServer.call(__MODULE__, {:unsubscribe, pid})
  end

end
