defmodule Reactivity.Signal.Derived do
  
  @moduledoc """
  A derived signal has at least one dependee.
  """
  use GenServer
  
  require Logger
  
  alias Reactivity.Signal.Derived
  alias Reactivity.Signal.Generic

  # last_value: The last generated value of this node.
  # generator : The lambda that will generate the next value (called with last value and value of parents).
  # children  : The children that depend on this node.
  # parent    : 
  # parents   : %{{:signal, type, pid}, idx}
  # lasts     : %{idx, value}
  defstruct last_value: nil, generator: nil, children: [], parents: %{}, lasts: %{}

  #############
  # GenServer #
  #############

  def start_link(args) do
    GenServer.start_link(__MODULE__, args, name: __MODULE__)
  end

  def init(%{val: v, gen_func: f, parents: parents}) do
    # For each dependee we need the last value to initialize our lasts map.
    last_values = parents
    |> Enum.map(fn({node, id}) ->
      # Get last value
      last = Generic.last_value(node)
      {id, last}
    end)
    |> Map.new
    
    {:ok, %Derived{last_value: v, generator: f, children: [], parents: parents, lasts: last_values}}
  end

  #############
  # Callbacks #
  #############

  @doc """ 
  Adds a child to the node. Whenever this node receives a new value it
  will "tick" all its children with the new value.  
  """
  def handle_call({:add_child, node}, _from, state) do
    new_state = %{state | children: [node | state.children]}
    {:reply, :ok, new_state}
  end

  @doc """
  Returns the last value generated by this node.
  """
  def handle_call({:last_value}, _from, state) do
    {:reply, {:ok, state.last_value}, state}
  end
  
  @doc """ 
  Is called when the parent has sent us a new value. If this happens we need to
  contact all our children as well withour newly computed value.
  """
  def handle_info({:tick, new_value, parent}, state) do

    # Compute the dependency index.
    idx = Map.get(state.parents, parent)

    # Update the last values.
    new_lasts = Map.put(state.lasts, idx, new_value)
    
    Logger.debug "#{inspect new_lasts}"

    # Compute this node's new value.
    new_value = apply(state.generator, Map.values(new_lasts))


    # Tick all my children.
    Enum.map(state.children, fn(c) -> send(c, {:tick, new_value, {:signal, :derived, self()}}) end)

    # Compute the new state.
    new_state = %{state | last_value: new_value, lasts: new_lasts}

    {:noreply, new_state}
  end
  
  #############
  # Interface #
  #############

  @doc """
  Creates a new source signal based on a function.
  It's not the point to be used in programs, merely as a simulation.
  """
  def new(gen_func, default, parents) do
    {:ok, pid} = GenServer.start_link(__MODULE__, %{val: default, gen_func: gen_func, parents: parents})
    {:ok, {:signal, :derived, pid}}
  end


end
